﻿using System;
using System.Windows;
using System.Collections.Generic;
using System.Windows.Controls;
using System.IO;
using System.Windows.Media; // VisualTreeHelper
using System.Collections.ObjectModel;
using System.Diagnostics;

namespace emanual.Wpf.Controls
{
	internal class TreeViewHelper
	{
		// XML ファイルからツリー構造を定義するデータを読み込んでツリー構造を構築する
		// treeView : 操作対象の TreeView
		// fileName : XML ファイル名
		// xpath    : XPath 式
		public static void LoadFromXmlFile(TreeView treeView, string fileName, string xpath)
		{
			System.Xml.XmlDocument document = null;
			System.Windows.Data.XmlDataProvider provider = null;
			System.Windows.Data.Binding binding = null;

			try
			{
				document = new System.Xml.XmlDocument();
				document.Load(fileName);

				provider = new System.Windows.Data.XmlDataProvider();
				provider.Document = document;

				binding = new System.Windows.Data.Binding();
				binding.Source = provider;
				binding.XPath = xpath;

				treeView.SetBinding(TreeView.ItemsSourceProperty, binding);
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message);
			}
		}

		//---------------------------------------------------------------------------------------------
		// ツリー構造を定義するファイルからデータを読み込んでツリー構造を構築する
		// ファイルの内容は、ノードのテキストを [Tab] で階層化したもので、次のとおり。
		//   東京都
		//   [Tab] 港区
		//   [Tab}[Tab] 赤坂
		// treeView : 操作対象の TreeView
		// fileName : ファイル名
		// encoding : エンコーディング法
		public static void LoadFromFile(TreeView treeView, string fileName, System.Text.Encoding encoding)
		{
			FileStream stream = null;

			try
			{
				stream = new FileStream(fileName, FileMode.Open, FileAccess.Read);
				LoadFromStream(treeView, stream, encoding);
			}
			catch (Exception ex)
			{
				throw new Exception(ex.Message);
			}
			finally
			{
				if (stream != null)
					stream.Close();
			}
		}

		//---------------------------------------------------------------------------------------------
		// ストリームからツリー構造を構築する（ストリームの内容は LoadFromFile メソッドにならう）
		public static void LoadFromStream(TreeView treeView, Stream stream, System.Text.Encoding encoding)
		{
			int count = 0;
			StreamReader reader = null;
			List<string> list = null;

			try
			{
				reader = new StreamReader(stream, encoding);
				list = new List<string>();

				while (reader.Peek() > -1)
				{
					list.Add(reader.ReadLine());
					++count;
				}
			}
			catch (Exception ex)
			{
				throw new Exception(ex.Message);
			}
			finally
			{
				if (reader != null)
					reader.Close();
			}

			string[] array = new string[count];

			list.CopyTo(array);

			LoadFromStringArray(treeView, array);
		}

		//---------------------------------------------------------------------------------------------
		// 文字列配列からツリー構造を構築する
		public static void LoadFromStringArray(TreeView treeView, string[] array)
		{
			TreeItemData currentItem = null, newItem = null, parentItem = null;
			int level;
			string text;

			for (int i = 0; i < array.Length; i++)
			{
				text = array[i].Trim(); // 前後のホワイトスペースを削除する

				level = 0;
				for (int j = 0; j < array[i].Length; j++)
				{
					if (array[i][j] == '\t') // 先頭の \t の数をかぞえる
						++level;
					else
						break;
				}

				//---------------------------------------------------
				if (currentItem == null) // ルートノードのとき
				{
					newItem = new TreeItemData(text, 0, null);
					treeView.Items.Add(newItem);
				}
				else if (currentItem.Level == level) // 直前のノードと同じレベルのとき
				{
					if (currentItem.Parent == null)
					{
						newItem = new TreeItemData(text, level, null);
						treeView.Items.Add(newItem);
					}
					else
					{
						newItem = new TreeItemData(text, level, currentItem.Parent);
						currentItem.Parent.Items.Add(newItem);
					}
				}
				else if (currentItem.Level == level - 1) // 直前のノードより 1 つ下のレベルのとき
				{
					newItem = new TreeItemData(text, level, currentItem);
					currentItem.Items.Add(newItem);
				}
				else if (currentItem.Level > level) // 直前のノードより上のレベルのとき
				{
					parentItem = currentItem.Parent;

					while (parentItem.Level > level)
					{
						parentItem = parentItem.Parent;
					}

					if (parentItem.Parent == null)
					{
						newItem = new TreeItemData(text, 0, null);
						treeView.Items.Add(newItem);
					}
					else
					{
						newItem = new TreeItemData(text, parentItem.Level, parentItem.Parent);
						parentItem.Parent.Items.Add(newItem);
					}
				}
				else
					throw new Exception("ツリー構造の構築に失敗しました");

				currentItem = newItem;
			} // end of for index
		}

		//---------------------------------------------------------------------------------------------
		// 文字列配列からツリー構造を構築する（TreeViewItem の Tag プロパティにオブジェクトを設定可能）
		// このメソッドの場合、list は実状に合わせて変更すること
		public static void LoadFromStringArray(TreeView treeView, string[] array, List<object>  list)
		{
			TreeItemData currentItem = null, newItem = null, parentItem = null;
			int level;
			string text;

			for (int i = 0; i < array.Length; i++)
			{
				text = array[i].Trim(); // 前後のホワイトスペースを削除する

				level = 0;
				for (int j = 0; j < array[i].Length; j++)
				{
					if (array[i][j] == '\t') // 先頭の \t の数をかぞえる
						++level;
					else
						break;
				}

				//---------------------------------------------------
				if (currentItem == null) // ルートノードのとき
				{
					newItem = new TreeItemData(text, 0, null, list[i]);
					treeView.Items.Add(newItem);
				}
				else if (currentItem.Level == level) // 直前のノードと同じレベルのとき
				{
					if (currentItem.Parent == null)
					{
						newItem = new TreeItemData(text, level, null, list[i]);
						treeView.Items.Add(newItem);
					}
					else
					{
						newItem = new TreeItemData(text, level, currentItem.Parent, list[i]);
						currentItem.Parent.Items.Add(newItem);
					}
				}
				else if (currentItem.Level == level - 1) // 直前のノードより 1 つ下のレベルのとき
				{
					newItem = new TreeItemData(text, level, currentItem, list[i]);
					currentItem.Items.Add(newItem);
				}
				else if (currentItem.Level > level) // 直前のノードより上のレベルのとき
				{
					parentItem = currentItem.Parent;

					while (parentItem.Level > level)
					{
						parentItem = parentItem.Parent;
					}

					if (parentItem.Parent == null)
					{
						newItem = new TreeItemData(text, 0, null, list[i]);
						treeView.Items.Add(newItem);
					}
					else
					{
						newItem = new TreeItemData(text, parentItem.Level, parentItem.Parent, list[i]);
						parentItem.Parent.Items.Add(newItem);
					}
				}
				else
					throw new Exception("ツリー構造の構築に失敗しました");

				currentItem = newItem;
			} // end of for index
		}

		//*********************************************************************************************
		// リスト項目の操作
		//*********************************************************************************************
		// 指定の項目から検索を開始し、表示状態にある直前の項目を取得する（閉じている項目は検索対象外）
		// item : 検索を開始する項目
		public static ItemsControl GetPrevVisibleItem(TreeViewItem item)
		{
			ItemsControl parentItemsControl = ItemsControl.ItemsControlFromItemContainer(item);

			if (parentItemsControl != null)
			{
				int index = parentItemsControl.ItemContainerGenerator.IndexFromContainer(item);

				while (index > 0)
				{
					index--;
					TreeViewItem treeViewItem = parentItemsControl.ItemContainerGenerator.ContainerFromIndex(index) as TreeViewItem;

					if (treeViewItem != null && treeViewItem.IsEnabled)
					{
						TreeViewItem treeViewItem2 = GetLastVisibleItem(treeViewItem);
						if (treeViewItem2 != null)
						{
							return treeViewItem2;
						}
					}
				}

				return parentItemsControl;
			}

			return null;
		}

		//------------------------------------------------------------------------------
		// GetPrevVisibleItem メソッドから呼び出されて、最後の見えている項目を取得する
		private static TreeViewItem GetLastVisibleItem(TreeViewItem item)
		{
			int count = -1;
			TreeViewItem treeViewItem = null;
			TreeViewItem treeViewItem2 = null;

			while (item != null)
			{
				if (item.IsEnabled)
				{
					if (!item.IsExpanded || !item.HasItems)
					{
						return item;
					}

					treeViewItem = item;
					treeViewItem2 = item;
					count = item.Items.Count - 1;
				}
				else
				{
					if (count <= 0)
					{
						break;
					}
					count--;
				}

				item = treeViewItem2.ItemContainerGenerator.ContainerFromIndex(count) as TreeViewItem;
			}

			if (treeViewItem != null)
			{
				return treeViewItem;
			}

			return null;
		}

		//---------------------------------------------------------------------------------------------
		// 指定の項目から検索を開始し、表示状態にある次の項目を取得する（閉じている項目は検索対象外）
		// isSearchSubtree : サブツリーを検索するとき true（最初の呼び出し時は常に true）
		// item : 検索を開始する項目
		public static TreeViewItem GetNextVisibleItem(bool isSearchSubtree, TreeViewItem item)
		{
			if (isSearchSubtree && item.IsExpanded && item.HasItems) // 子項目を持ち、開いているとき
			{
				TreeViewItem treeViewItem = item.ItemContainerGenerator.ContainerFromIndex(0) as TreeViewItem;

				if (treeViewItem != null)
				{
					if (treeViewItem.IsEnabled)
					{
						return treeViewItem;
					}

					return GetNextVisibleItem(false, treeViewItem);
				}
			}

			ItemsControl parentItemsControl = ItemsControl.ItemsControlFromItemContainer(item);

			if (parentItemsControl != null)
			{
				int index = parentItemsControl.ItemContainerGenerator.IndexFromContainer(item);
				int count = parentItemsControl.Items.Count;
				TreeViewItem treeViewItem2;

				while (index < count)
				{
					index++;
					treeViewItem2 = parentItemsControl.ItemContainerGenerator.ContainerFromIndex(index) as TreeViewItem;
					if (treeViewItem2 != null && treeViewItem2.IsEnabled)
					{
						return treeViewItem2;
					}
				}

				treeViewItem2 = parentItemsControl as TreeViewItem;

				if (treeViewItem2 != null)
				{
					return GetNextVisibleItem(false, treeViewItem2);
				}
			}

			return null;
		}

		//---------------------------------------------------------------------------------------------
		// すべてのノードのリストを取得する
		public static List<TreeItemData> GetAllItems(TreeView treeView)
		{
			var list = new List<TreeItemData>();

			for (int i = 0; i < treeView.Items.Count; ++i)
			{
				var item = treeView.Items[i] as TreeItemData;

				if (item != null)
				{
					var data = item.Clone();
					data.Level = 0;

					list.Add(data);

					if (item.HasItems)
					{
						DoGetItems(list, item, 0);
					}
				}
			}

			return list;
		}

		//---------------------------------------------------------------------------------------------
		// GetTreeViewItems メソッドから呼び出される再帰メソッド
		private static void DoGetItems(IList<TreeItemData> list, TreeItemData parentItem, int level)
		{
			++level;
		
			for (int i = 0; i < parentItem.Items.Count; ++i)
			{
				var item = parentItem.Items[i] as TreeItemData;

				if (item != null)
				{
					var data = item.Clone();
					data.Level = level;

					list.Add(item);

					if (item.HasItems)
					{
						DoGetItems(list, item, level);
					}
				}
			}
		}

		//---------------------------------------------------------------------------------------------
		// 指定の項目のレベル（深度）を取得する（トップレベルは 0）
		// item : レベルを取得する項目
		public static int GetLevel(TreeViewItem item)
		{
			int level = 0;
			TreeViewItem parent = GetParentItem(item);

			while (parent != null)
			{
				++level;
				parent = GetParentItem(parent);
			}

			return level;
		}

		//---------------------------------------------------------------------------------------------
		// 指定の項目を左へ移動する（左へ移動とは上の階層への移動である）
		// obj : 移動する項目を指定するオブジェクト
		public static void MoveLeft(TreeView treeView, TreeViewItem item)
		{
			if (item == null)
				return;

			var parent = GetParentItem(item);

			// Parent を 持たないとき（ルートノードのとき）
			if (parent == null)
				return;

			// parent の次の項目として挿入する
			Remove(treeView, item);
			InsertAfter(treeView, parent, item);
			//item.IsSelected = true;
		}

		//---------------------------------------------------------------------------------------------
		// 指定の項目を右へ移動する（右へ移動とは下の階層への移動である）
		// obj : 移動する項目を指定するオブジェクト
		public static void MoveRight(TreeView treeView, TreeViewItem item)
		{
			if (item == null)
				return;

			var prevItem = GetPrevItem(treeView, item);

			if (prevItem == null) // 前の兄弟ノードがないので、何もしない
				return;
			else // 前の兄弟ノードの子ノードにする
			{
				Remove(treeView, item);
				prevItem.Items.Add(item);
			}

			item.IsSelected = true;
		}

		//---------------------------------------------------------------------------------------------
		// 指定の項目を上へ移動する（上へ移動とは同じレベル内における上方向への移動である）
		// obj : 移動する項目を指定するオブジェクト
		public static void MoveUp(TreeView treeView, TreeViewItem item)
		{
			if (item == null)
				return;

			int index = GetIndexFromItem(treeView, item);
			var parent = GetParentItem(item);
			var prevItem = (TreeViewItem)GetPrevItem(treeView, item);

			if (parent == null) // ルートノードのとき
			{
				if (prevItem != null) // 同じレベルで上の項目があるとき、順番を入れ替える
				{
					treeView.Items.Remove(item);
					treeView.Items.Insert(index - 1, item);
				}
			}
			else
			{
				if (prevItem != null) // 同じレベルで上の項目があるとき、順番を入れ替える
				{
					parent.Items.Remove(item);
					parent.Items.Insert(index - 1, item);
				}
			}

			item.IsSelected = true;
		}

		//---------------------------------------------------------------------------------------------
		// 指定の項目を下へ移動する（下へ移動とは同じレベル内における下方向への移動である）
		// obj : 下へ移動する項目を指定するオブジェクト
		public static void MoveDown(TreeView treeView, TreeViewItem item)
		{
			if (item == null)
				return;

			int index = GetIndexFromItem(treeView, item);
			var parent = GetParentItem(item);
			var nextItem = GetNextItem(treeView, item);

			if (parent == null) // ルートノードのとき
			{
				if (nextItem != null) // 同じレベルで下の項目があるとき、順番を入れ替える
				{
					treeView.Items.Remove(item);
					treeView.Items.Insert(index + 1, item);
				}
			}
			else
			{
				if (nextItem != null) // 同じレベルで下の項目があるとき、順番を入れ替える
				{
					parent.Items.Remove(item);
					parent.Items.Insert(index + 1, item);
				}
			}

			item.IsSelected = true;
		}

		//---------------------------------------------------------------------------------------------
		// 指定の項目を削除する
		// item : 削除する項目
		public static void Remove(TreeView treeView, TreeViewItem item)
		{
			if (item == null)
				return;

			var parent = GetParentItem(item);

			if (parent != null)
			{
				parent.Items.Remove(item);
			}
			else // ルートノードのとき
			{
				treeView.Items.Remove(item);
			}
		}

		//---------------------------------------------------------------------------------------------
		// 指定の項目を挿入または追加する
		// selected : 挿入位置の基準とする項目（この項目の前に挿入する）
		// item     : 挿入する項目
		public static void InsertBefore(TreeView treeView, TreeViewItem selected, TreeViewItem item)
		{
			if ((selected == null) || (item == null))
				return;

			int index = GetIndexFromItem(treeView, selected);
			var item1 = selected as TreeViewItem;
			var parent = GetParentItem(item1);

			if (parent == null)
				treeView.Items.Insert(index, item);
			else
				parent.Items.Insert(index, item);

			item.IsSelected = true;
		}

		//---------------------------------------------------------------------------------------------
		// 指定の項目を挿入または追加する
		// selected : 挿入位置の基準とする項目（この項目の次に挿入する）
		// item     : 挿入する項目
		public static void InsertAfter(TreeView treeView, TreeViewItem selected, TreeViewItem item)
		{
			if ((selected == null) || (item == null))
				return;

			int index = GetIndexFromItem(treeView, selected);
			var item1 = selected as TreeViewItem;
			var parent = GetParentItem(item1);

			if (parent == null) // ルートノードのとき
			{
				if (treeView.Items.Count - 1 > index)
					treeView.Items.Insert(index + 1, item);
				else
					treeView.Items.Add(item);
			}
			else
			{
				if (parent.Items.Count - 1 > index)
					parent.Items.Insert(index + 1, item);
				else
					parent.Items.Add(item);
			}

			item.IsSelected = true;
		}

		//---------------------------------------------------------------------------------------------
		// 指定の項目の同一レベル内におけるインデックスを取得する（先頭は 0）
		// item : 検索する項目を指定するオブジェクト（選択状態の項目）
		public static int GetIndexFromItem(TreeView treeView, TreeViewItem item)
		{
			int result = -1;

			if (item == null)
				return -1;

			var parent = GetParentItem(item);

			if (parent != null)
			{
				for (int i = 0; i < parent.Items.Count; i++)
				{
					var item1 = parent.Items[i] as TreeViewItem;

					if (item1.Equals(item))
					{
						result = i;
						break;
					}
				}
			}
			else
			{
				for (int i = 0; i < treeView.Items.Count; i++)
				{
					var item1 = treeView.Items[i] as TreeViewItem;

					if (item1.Equals(item))
					{
						result = i;
						break;
					}
				}
			}

			return result;
		}

		//---------------------------------------------------------------------------------------------
		// すべてのノードを閉じる
		public static void CollapseAll(TreeView treeView)
		{
			for (int i = 0; i < treeView.Items.Count; i++)
			{
				var item = treeView.Items[i] as TreeViewItem; //<<

				if (item != null)
				{
					item.IsExpanded = false;
				}
			}

			for (int i = 0; i < treeView.Items.Count; i++)
			{
				DoCollapseAll((TreeViewItem)(treeView.Items[i]));
			}

			var item2 = treeView.Items[0] as TreeViewItem;
			item2.IsSelected = true;
		}

		//---------------------------------------------------------------------------------------------
		// CollapseAll メソッドの再帰的に処理するメソッド
		private static void DoCollapseAll(TreeViewItem parentItem)
		{
			for (int i = 0; i < parentItem.Items.Count; i++)
			{
				var item = parentItem.Items[i] as TreeViewItem; //<<

				if (item != null)
				{
					item.IsExpanded = false;

					if (item.HasItems)
					{
						DoCollapseAll(item);
					}
				}
			}
		}

		//---------------------------------------------------------------------------------------------
		// すべてのノードを展開する
		public static void ExpandAll(TreeView treeView)
		{
			for (int i = 0; i < treeView.Items.Count; ++i)
			{
				var item = treeView.Items[i] as TreeViewItem;

				if (item != null)
				{
					item.IsExpanded = true;
				}
			}

			for (int i = 0; i < treeView.Items.Count; i++)
			{
				DoExpandAll((TreeViewItem)(treeView.Items[i]));
			}
		}

		//---------------------------------------------------------------------------------------------
		// ExpandAll メソッドの再帰的に処理するメソッド
		private static void DoExpandAll(TreeViewItem parentItem)
		{
			for (int i = 0; i < parentItem.Items.Count; ++i)
			{
				var item = parentItem.Items[i] as TreeViewItem;

				if (item != null)
				{
					item.IsExpanded = true;

					if (item.HasItems)
					{
						DoExpandAll(item);
					}
				}
			}
		}

		//---------------------------------------------------------------------------------------------
		// 指定の項目と同じレベル内の最初の項目を取得する
		// item : 項目を指定するオブジェクト（TreeViewItem オブジェクト）
		public static TreeViewItem GetFirstItem(TreeView treeView, TreeViewItem item)
		{
			TreeViewItem result = null;

			if (item != null)
			{
				var parent = GetParentItem(item);

				if (parent != null)
					result = parent.Items[0] as TreeViewItem;
				else
					result = treeView.Items[0] as TreeViewItem;
			}

			return result;
		}

		//---------------------------------------------------------------------------------------------
		// 指定の項目と同じレベル内の最後の項目を取得する
		// item : 項目を指定するオブジェクト（TreeViewItem オブジェクト）
		public static TreeViewItem GetLastItem(TreeView treeView, TreeViewItem item)
		{
			TreeViewItem result = null;

			if (item != null)
			{
				var parent = GetParentItem(item);

				if (parent != null)
					result = parent.Items[parent.Items.Count - 1] as TreeViewItem;
				else
					result = treeView.Items[treeView.Items.Count - 1] as TreeViewItem;
			}

			return result;
		}

		//---------------------------------------------------------------------------------------------
		// 指定の項目の同じレベル内の次の項目を取得する（次の項目がないとき、null を返すことに注意）
		// item : 項目を指定するオブジェクト（TreeViewItem オブジェクト）
		public static object GetNextItem(TreeView treeView, TreeViewItem item)
		{
			TreeViewItem result = null;

			if (item != null)
			{
				int index = GetIndexFromItem(treeView, item);
				var parent = GetParentItem(item);

				if (parent == null) // ルートノードのとき
				{
					if (treeView.Items.Count - 1 > index)
						result = (TreeViewItem)treeView.Items[index + 1] as TreeViewItem;
					else
						result = null;
				}
				else
				{
					if (parent.Items.Count - 1 > index)
						result = (TreeViewItem)parent.Items[index + 1] as TreeViewItem;
					else
						result = null;
				}
			}

			return result;
		}

		//---------------------------------------------------------------------------------------------
		// 指定の項目の同じレベル内の前の項目を取得する（前の項目がないとき、null を返すことに注意）
		// item : 項目を指定するオブジェクト（TreeViewItem オブジェクト）
		public static TreeViewItem GetPrevItem(TreeView treeView, TreeViewItem item)
		{
			TreeViewItem result = null;

			if (item != null)
			{
				int index = GetIndexFromItem(treeView, item);
				var parent = GetParentItem(item);

				if (parent == null) // ルートノードのとき
				{
					if (index > 0)
						result = (TreeViewItem)treeView.Items[index - 1] as TreeViewItem;
					else
						result = null;
				}
				else
				{
					if (index > 0)
						result = (TreeViewItem)parent.Items[index - 1] as TreeViewItem;
					else
						result = null;
				}
			}

			return result;
		}

		//---------------------------------------------------------------------------------------------
		// 指定のオブジェクトの項目を展開する
		// item : 展開する項目
		public static void Expand(TreeView treeView, TreeViewItem item)
		{
			if (item != null)
			{
				item.IsExpanded = true;
			}
		}

		//---------------------------------------------------------------------------------------------
		// 指定のオブジェクトの項目を閉じる
		// item : 閉じる項目
		public static void Collapse(TreeView treeView, TreeViewItem item)
		{
			if (item != null)
			{
				item.IsExpanded = false;
			}
		}

		//---------------------------------------------------------------------------------------------
		// 指定の TreeViewItem オブジェクトの親オブジェクトを取得する
		// item : TreeViewItem オブジェクト
		public static TreeViewItem GetParentItem(TreeViewItem item)
		{
			var parent = VisualTreeHelper.GetParent(item);
			var treeViewItem = parent as TreeViewItem;

			// null でない Visual オブジェクトまで検索する
			while (treeViewItem == null)
			{
				if (parent == null)
					break;

				parent = VisualTreeHelper.GetParent(parent);
				treeViewItem = parent as TreeViewItem;
			}

			return treeViewItem;
		}

		//---------------------------------------------------------------------------------------------
		// 指定のオブジェクトの項目を取得する
		// item : 検索する項目
		private static TreeViewItem GetItemFromObject(TreeView treeView, object obj)
		{
			TreeViewItem item = null;
			DependencyObject container = null;

			try
			{
				container = GetItemContainer(treeView, obj);
				var treeViewItem = container as TreeViewItem;

				while (treeViewItem == null)
				{
					container = VisualTreeHelper.GetParent(container);
					treeViewItem = container as TreeViewItem;
				}

				item = treeViewItem;
			}
			catch (Exception e)
			{
				MessageBox.Show("GetItemFromObject: " + e.Message);
			}

			return item;
		}

		//---------------------------------------------------------------------------------------------
		// 指定のオブジェクトのコンテナオブジェクトを取得する
		// obj : 検索する項目を指定するオブジェクト
		public static DependencyObject GetItemContainer(TreeView treeView, object obj)
		{
			DependencyObject container = treeView.ItemContainerGenerator.ContainerFromItem(obj);

			// container には通常 TreeViewItem オブジェクトが戻るはずだが、null が戻る可能性がある
			//そこで、 container が null のとき、Visual ツリーをさかのぼる
			var element = obj as FrameworkElement;

			while ((container == null) && (element != null))
			{
				element = VisualTreeHelper.GetParent(element) as FrameworkElement;
				container = element as TreeViewItem;
			}

			return container;
		}
	} // end of TreeViewHelper class

	//*********************************************************************************************
	// TreeItemData クラス（LoadFromFile および LoadFromStream メソッドで使うデータ構造）
	//*********************************************************************************************
	public class TreeItemData : TreeViewItem
	{
		public int Level { get; set; }
		public new TreeItemData Parent { get; set; }

		public TreeItemData(string header)
		{
			this.Header = header;
		}

		public TreeItemData(string header, int level, TreeItemData parent)
		{
			this.Header = header;
			this.Level = level;
			this.Parent = parent;
		}

		public TreeItemData(string header, int level, TreeItemData parent, object tag)
		{
			this.Header = header;
			this.Level = level;
			this.Parent = parent;
			this.Tag = tag;
		}

		public TreeItemData Clone()
		{
			return (TreeItemData)this.MemberwiseClone();
		}
	} // end of TreeItemData class
} // end of emanual.Wpf.Utility namespace
